## Copyright (C) 2026 Andreas Bertsatos <abertsatos@biol.uoa.gr>
##
## This file is part of the statistics package for GNU Octave.
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.
##
##
## -*- texinfo -*-
## @deftypefn {Private Function} {@var{tf} =} classMembers (@var{Y}, @var{C}, @var{i})
## Which observations of a response belong to the @var{i}th class.
##
## The complement of @code{labelIndex}: that asks which class an observation
## is in, this asks which observations are in a class.  A response naming its
## classes in the rows of a character matrix is matched as whole names, with
## the padding such a matrix carries stripped from both sides.
## @end deftypefn

function tf = classMembers (Y, C, i)

  if (ischar (C) || ischar (Y))
    tf = strcmp (cellstr (Y), strtrim (char (labelsFromIndex (C, i))));
  else
    tf = ismember (Y, C(i));
  endif

endfunction
