## Copyright (C) 2026 Andreas Bertsatos <abertsatos@biol.uoa.gr>
##
## This file is part of the statistics package for GNU Octave.
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Private Function} {@var{str} =} classNameListing (@var{ClassNames})
##
## The class names as a model's display lists them, quoted when they are
## text and braced as a set.
##
## @end deftypefn

## The classes as the display lists them, quoted when they are text.
function str = classNameListing (ClassNames)

  if (iscellstr (ClassNames))
    fmt = repmat ({'''%s'''}, 1, numel (ClassNames));
    fmt = strjoin (fmt, '  ');
    str = sprintf (fmt, ClassNames{:});
  elseif (ischar (ClassNames))
    cn = cellstr (ClassNames);
    fmt = repmat ({'''%s'''}, 1, numel (cn));
    fmt = strjoin (fmt, '  ');
    str = sprintf (fmt, cn{:});
  else
    fmt = repmat ({'%g'}, 1, numel (ClassNames));
    fmt = strjoin (fmt, '  ');
    str = sprintf (fmt, ClassNames);
  endif
  str = ['{', str, '}'];

endfunction
